#!/usr/bin/env python
#*****************************************************************************#
#**
#**  Multiview Training and Testing Framework
#**
#**    Kyle D. Feuz, 2014
#**
#** Tools by the Center for Advanced Studies in Adaptive Systems at
#**  the School of Electrical Engineering and Computer Science at
#**  Washington State University
#** 
#** Copyright Washington State University, 2014
#** Copyright Kyle D. Feuz, 2014
#** 
#** Licenced under the Creative Commons (http://creativecommons.org):
#** Attribution-NonCommercial-ShareAlike 3.0 Unported
#** 
#** If this code is used for public research, any resulting publications need
#** to cite work done by Kyle D. Feuz at the Center for Advanced Study of 
#** Adaptive Systems (CASAS) at Washington State University.
#** 
#** Contact: Kyle D. Feuz (kyle.feuz@wsu.edu)
#** Contact: Diane J. Cook (cook@eecs.wsu.edu)
#*****************************************************************************#
import utils
#from lpp import Lpp
from Orange.data import sample, Table
from Orange.projection.linear import Pca
import numpy as np
import sys
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
import logging.handlers

LOG_FILENAME = 'multiview.log'
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
#Rotate log every 10 mb keep the last ten logs
handler = logging.handlers.RotatingFileHandler(LOG_FILENAME, maxBytes=10000000, backupCount=10)
formatter = logging.Formatter("%(asctime)s --%(name)s-- %(message)s")
handler.setFormatter(formatter)
ch = logging.StreamHandler()
ch.setLevel(logging.WARN)
ch.setFormatter(formatter)
logger.addHandler(ch)

v_start = [0,36,133,230,0]
v_end= [36,133,230,243,243]
class_vars = range(243,250)

class Data:

   def __init__(self):
      self.algs = []
      self.training = []
      self.labeled = []
      self.unlabeled = []
      self.validation = []
      self.class_labels = []
      self.bootstrapped_labels = []
      self.indices = []
      self.results_dict = {}

   def reset_labels(self, keep_bootstrapped=True):
      for view in self.training:
         for d in view:
            d.set_class(self.class_labels[int(d[utils._ID])])
      if self.bootstrapped_labels and keep_bootstrapped:
         for view in self.training[1:]:
            bootstrapped = view.select_ref(self.indices, 1)
            for i,d in enumerate(bootstrapped):
               d.set_class(self.bootstrapped_labels[i])
   
def baseline(data, train, show_cm=False):
   logger.info("Baseline results")
   rd = utils.train_test_report(data.algs, train, data.validation, show_cm)
   return rd

def cotrain(data, show_cm=False):
   labeled = data.labeled
   unlabeled = data.unlabeled
   teach_ind = data.teach_ind[:]
   learn_ind = data.learn_ind[:]
   

   Nsubset = sample.SubsetIndices2(p0=len(data.training)*data.N)
   Nsubset.stratified = True
   half = sample.SubsetIndicesCV(folds=len(data.training))
   half.stratified = True

   iteration = 0
   while 1:
      if iteration > 0:
         # Re-sort the data since new points have been added
         # Keep the bootstrapped labels properly placed
         labeled = []
         unlabeled = []
         labeled.append(data.training[0].select_ref(teach_ind, 0))
         unlabeled.append(data.training[0].select_ref(teach_ind, 1))
         for view in data.training[1:]:
            labeled.append(view.select_ref(learn_ind, 0))
            unlabeled.append(view.select_ref(learn_ind, 1))
         if len(unlabeled[0]) == 0:
            break
         logger.debug("Labeled Size {0}".format(len(labeled[0])))
         logger.debug('Unlabeled Size {0}'.format(len(unlabeled[0])))
    
      #Build classifiers
      classifiers = utils.train(data.algs[0:1], labeled)
      if((iteration % 10) == 1):
         results = utils.test(classifiers, data.validation)
         logger.info("Co-training results after {0} iterations".format(iteration))
         utils.show_results(results, data.algs[0:1], show_cm)
      
      # Select subset to label
      # Split test in half
      logger.debug("Sampling the unlabeled data")
      subset = []
      if len(unlabeled[0]) > len(data.training)*data.N:
         for uview in unlabeled:
            Nsubset_ind = Nsubset(unlabeled[0])
            subset.append(uview.select_ref(Nsubset_ind, 0))
      else:
         for uview in unlabeled:
            subset.append(uview)
      half_ind = half(subset[0])
      t = []
      for i,sub in enumerate(subset):
         t.append(sub.select_ref(half_ind, i))
      

      # Label subset
      for i,classifier in enumerate(classifiers):
         logger.debug("Labeling sampled data")
         labels = []
         for d in t[i]:
            c,prob = classifier[0](d, 2) #2 = GetBoth 
            labels.append((c, prob[c], int(d[utils._ID]), d.get_class()))
         # Select top n from each class
         logger.debug("Choosing top n")
         labels.sort(reverse=True)
         count = 0
         value = 100
         correct = 0
         total = 0.0
         cm = [[0 for y in xrange(len(t[i].domain.class_var.values)+1)] for x in xrange(len(t[i].domain.class_var.values)+1)]
         for label in labels:
            predicted = int(label[0])
            if predicted > value:
               continue
            while predicted < value:
               value -=1
            count += 1
            if count == data.n:
               count = 0
               value -=1
            actual = int(data.training[0][label[2]].get_class())
            cm[actual][predicted] += 1
            if actual == predicted:
               correct += 1
            for view in data.training:
               view[label[2]].set_class(label[0])
            # Set index for labeled data
            teach_ind[label[2]]=0
            learn_ind[label[2]]=0
            total += 1
         if total == 0.0:
            total = 0.1
         logger.info("Correct: {0} Accuracy: {0}/{1} {2}".format(correct, total, correct/total))
         if show_cm:
            for row in cm:
               logger.info('{0}'.format(row))
         # Repeat
      iteration += 1


   #Baseline validation results
   logger.info("Co-Training results after {0} iterations".format(iteration))
   training = data.training[:]
   if data.bootstrapped_labels:
      training[0] = training[0].select_ref(data.indices, 1, negate=True)
   rd = utils.train_test_report(data.algs, training,
                                data.validation, show_cm)
   return rd

def coem(data, show_cm=False):

   labeled = data.labeled
   unlabeled = data.unlabeled
   # Don't use teacher bootstrapped labels
   # On teacher view
   training = data.training[:]
   if data.bootstrapped_labels:
      training[0] = training[0].select_ref(data.indices, 1, negate=True)

   classifiers = utils.train(data.algs[0:1], data.labeled)
   classifier = classifiers[0][0]
   #Label the test data for view [0]
   for d in unlabeled[0]:
      c = classifier(d)
      data.training[1][int(d[utils._ID])].set_class(c)

   for i in xrange(data.n):
      logger.debug("Iterating  CoEM")
      for j, view in enumerate(training):
         if i == 0 and j == 0:
            continue
         #Build classifier for view j
         classifier = utils.train(data.algs[0:1], [view], j)[0][0]
         nextview = (j+1) % len(data.training)
         logger.debug("View {0} labeling View {1}".format(j, nextview))
         #Label the test data
         total = 0.0
         agreed = 0
         for d in unlabeled[j]:
            c = classifier(d)
            total += 1
            if int(c) == int(d.get_class()):
               agreed += 1
            data.training[nextview][int(d[utils._ID])].set_class(c)
         if total == 0.0:
            total = 0.1
         logger.info("Agreement: {0}/{1} {2}".format(agreed, total, agreed/total))
      if (i%5) == 0 and data.n > 5:
         logger.info("Co-EM results after {0} iterations".format(i+1))
         utils.train_test_report(data.algs, training,
                                 data.validation, show_cm)

   logger.info("Co-EM results after {0} iterations".format(data.n))
   rd = utils.train_test_report(data.algs, training,
                                data.validation, show_cm)
   return rd


def manalign(data, show_cm=False):
   logger.info("Projecting via PCA")
   pca = []
   projected = []
   min_size = 999999
   for view in data.training:
      size = len(view.domain.features)
      if size < min_size:
         min_size = size
   for view in data.unlabeled:
      pca.append(Pca(view, max_components=min_size))
      projected.append(pca[-1](view))

   logger.info("Translating projections")
   source = pca[0](data.labeled[0])
   x = projected[0].to_numpy()[0]
   targets = [pca[0](data.validation[0])]
   for i,view in enumerate(projected[1:]):
      y,c,w = projected[i+1].to_numpy()
      z = np.dot(y.T, x)
      u,s,v = np.linalg.svd(z)
      q = np.dot(u, v)
      k = sum(s)/np.trace( np.dot(y.T, y))

      #yprime = np.dot((k*y), q)
      #c = c.reshape(len(c),1)
      #targets.append(Table(source.domain, np.append(yprime, c, axis=1)))

      #Compute out-of-sample results
      vp = pca[i+1](data.validation[i+1])
      #translate target
      y,c,w = vp.to_numpy()
      yprime = np.dot((k*y), q)
      c = c.reshape(len(c),1)
      targets.append(Table(source.domain, np.append(yprime, c, axis=1)))

   #Results
   logger.info("Manifold Alignment results")
   train = [source for x in xrange(len(targets))]
   rd = utils.train_test_report(data.algs, train, targets, show_cm)
   return rd



def teach_learn(data, show_cm=False):
   labeled = data.labeled
   unlabeled = data.unlabeled
   #Build teacher classifier
   classifiers = utils.train(data.algs[0:1], labeled[0:1])
   
   #Save teacher classifier
   teacher = classifiers[0][0]
   
   #Use teacher to label data
   logger.debug("Labeling data for learner")
   num_classes = len(data.training[0].domain.class_var.values)+1
   cm = [[0 for y in xrange(num_classes)] for x in xrange(num_classes)]
   correct = 0
   total = 0.0
   labels = []
   for d in unlabeled[0]:
      c,prob = teacher(d, 2) # 2 = GetBoth
      #Comment out for 'most confident' labels
      predicted = int(c)
      actual = int(d.get_class())
      label = (c,prob[c],int(d[utils._ID]),d.get_class())

      #Uncomment to use 'most confident' labels
      #labels.append((c,prob[c],int(d[utils._ID]),d.get_class()))
   # Select top n from each class
   #labels.sort(reverse=True)
   #count = 0
   #value = 100
   #for label in labels:
   #   predicted = int(label[0])
   #   if predicted > value:
   #      continue
   #   while predicted < value:
   #      count = 0
   #      value -= 1
   #   count += 1
   #   if count == data.n*999999: #Total Hack.  Fix this
   #      count = 0
   #      value -= 1
   #   actual = int(label[3])
      cm[actual][predicted] += 1
      if actual == predicted:
         correct += 1
      for view in data.training[1:]:
         view[label[2]].set_class(label[0])
      total += 1
   if total == 0.0:
      total = 0.1
   logger.info("Correct:{0} Accuracy: {0}/{1} {2}".format(correct, total, correct/total))
   if show_cm:
      for row in cm:
         logger.info('{0}'.format(row))

   #Train learners on newly labeled data
   #Test on validation set
   logger.info("Teacher/Learner results")
   training = data.training[:]
   if data.bootstrapped_labels:
      training[0] = training[0].select_ref(data.indices, 1, negate=True)
   rd = utils.train_test_report(data.algs, training,
         data.validation, show_cm)
   return rd

   #Compare prediction against teacher
   #Not yet ready for multiple learners
   #correct = [[0 for x in range(len(data.algs))] for y in range(len(classifiers[1:]))]
   #for i in xrange(len(data.validation[0])):
   #   l_c = teacher(data.validation[0][i])
   #   for j,view in enumerate(classifiers[1:]):
   #      for k,learner in view:
   #         l_l = learner(data.validation[j+1][i])
   #         correct[j][k] += 1 if int(l_l) == int(l_c) else 0
   #total = float(len(data.validation[0]))
   #for i,view in enumerate(classifiers[1:]):
   #   for j,alg in enumerate(data.algs):
   #      print "View {0} {1} accuracy against teacher is {2}".format(i+1, alg, correct[i][j]/total)

def _selftrain(classifiers, data):
   logger.debug("Self-Labeling data")
   for i,classifier in enumerate(classifiers):
      teacher = classifier[0]
      #Use teacher to label data
      num_classes = len(data.training[i].domain.class_var.values)+1
      cm = [[0 for y in xrange(num_classes)] for x in xrange(num_classes)]
      correct = 0
      total = 0.0
      for d in data.unlabeled[i]:
         c,prob = teacher(d, 2) # 2 = GetBoth
         predicted = int(c)
         actual = int(d.get_class())
         cm[actual][predicted] += 1
         if actual == predicted:
            correct += 1
         d.set_class(c)

def selftrain(data, show_cm=False):
   #Build classifiers
   classifiers = utils.train(data.algs[0:1], data.labeled)
   _selftrain(classifiers, data)
   for i in xrange(data.n-1):
      classifiers = utils.train(data.algs[0:1], data.training)
      _selftrain(classifiers, data)
   

   logger.info("Self Training results")
   training = data.training[:]
   if data.bootstrapped_labels:
      training[0] = training[0].select_ref(data.indices, 1, negate=True)
   rd = utils.train_test_report(data.algs, training,
                                data.validation, show_cm)
   return rd

def evaluate(training, validation, args):

   modes = set(args.mode)

   class_labels = []
   bootstrap_labels = []
   for d in training[0]:
      class_labels.append(d.get_class())
   
   if args.teacher_split:
      logger.debug("Splitting Teacher/Learner Data")
      if args.teacher_split + args.split >= 1:
	 indicesN = sample.SubsetIndices2(p=args.teacher_split)
      else:
         indicesN = sample.SubsetIndicesN(p=[args.teacher_split, args.split])
      ind = indicesN(training[0])
      labeled = []
      unlabeled = []
      for i,view in enumerate(training):
         if i == 0:
            labeled.append(view.select_ref(ind, 2, negate=True))
         else:
            labeled.append(view.select_ref(ind, 1))
         unlabeled.append(view.select_ref(ind, 2))
      if args.bootstrap:
         logger.debug("Bootstrapping Data")
         #Don't use bootstrapped labels for the teacher
         labeled[0] = training[0].select_ref(ind, 0)
         bootstrap_test = training[0].select_ref(ind, 1)
         # We can't just filter these out though because it breaks the indexes
         #training[0] = training[0].select_ref(ind, 1, negate=True)
         teacher = utils.train(args.algorithms[:1], labeled[:1])[0][0]
         for d in bootstrap_test:
            c,prob = teacher(d, 2) # 2 = GetBoth
            bootstrap_labels.append(c)
            for view in training[1:]:
               view[int(d[utils._ID])].set_class(c)
   else:
      logger.debug("Splitting Labeled/Unlabeled data")
      indices2 = sample.SubsetIndices2(p0=args.split)
      indices2.stratified = True
      ind = indices2(training[0])
      labeled = []
      unlabeled = []
      for view in training:
         labeled.append(view.select_ref(ind, 0))
         unlabeled.append(view.select_ref(ind, 1))

   logger.debug("Teacher Labeled Size {0}".format(len(labeled[0])))
   logger.debug("Labeled(Bootstrapped) Size {0}".format(len(labeled[1])))
   logger.debug('Unlabeled Size {0}'.format(len(unlabeled[0])))
   logger.debug('Valdiation Size {0}'.format(len(validation[0])))

   if max(ind) == 2:
      teach_ind = []
      learn_ind = []
      if args.bootstrap:
         for i in ind:
            if i == 0:
               teach_ind.append(0)
               learn_ind.append(2)
            elif i == 1:
               teach_ind.append(2)
               learn_ind.append(0)
            else:
               teach_ind.append(1)
               learn_ind.append(1)
      else:
         for i in ind:
            if i == 0:
               teach_ind.append(0)
               learn_ind.append(2)
            elif i == 1:
               teach_ind.append(0)
               learn_ind.append(0)
            else:
               teach_ind.append(1)
               learn_ind.append(1)
   else:
      teach_ind = ind.clone()
      learn_ind = ind.clone()

   data = Data()
   data.algs = args.algorithms
   data.N = args.N
   data.n = args.n
   data.training = training
   data.labeled = labeled
   data.unlabeled = unlabeled
   data.validation = validation
   data.class_labels = class_labels
   data.bootstrapped_labels = bootstrap_labels
   data.indices = ind
   data.teach_ind = teach_ind
   data.learn_ind = learn_ind


   results_dict = {}
   if len(labeled[1]) < 1 or len(unlabeled[1]) < 1:
      logger.debug('Invalid Split {0}'.format(args.split))
      d = {y:{x:{'Acc':0.0, 'Recall':0.0} for x in data.algs} for y in xrange(len(training))}
      if 'baseline' in modes:
         results_dict['base_all'] = d 
         results_dict['base_labeled'] = d
         results_dict['base_unlabeled'] = d
      if 'maj' in modes:
         results_dict['majority'] = d
      if 'rand' in modes:
         results_dict['rand'] = d
      if 'cotrain' in modes:
         results_dict['cotrain'] = d
      if 'coem' in modes:
         results_dict['coem'] = d
      if 'manalign' in modes:
         results_dict['manalign'] = d
      if 'teach' in modes:
         results_dict['teach'] = d
      if args.bootstrap:
         data.reset_labels(False)
      return results_dict


   if 'baseline' in modes:
      training = data.training[:]
      if data.bootstrapped_labels:
         training[0] = training[0].select_ref(data.indices, 1, negate=True)
      results_dict['base_all'] = baseline(data, training)
      results_dict['base_labeled'] = baseline(data, data.labeled)
      results_dict['base_unlabeled'] = baseline(data, data.unlabeled)
   if 'maj' in modes:
      algs = data.algs
      data.algs = ['maj']
      results_dict['majority'] = baseline(data, data.labeled)
      data.algs = algs
   if 'rand' in modes:
      algs = data.algs
      data.algs = ['rand']
      results_dict['rand'] = baseline(data, data.labeled)
      data.algs = algs
   if 'cotrain' in modes:
      results_dict['cotrain'] = cotrain(data)
      data.reset_labels()
   if 'coem' in modes:
      results_dict['coem'] = coem(data)
      data.reset_labels()
   if 'manalign' in modes:
      results_dict['manalign'] = manalign(data)
   if 'teach' in modes:
      results_dict['teach'] = teach_learn(data)
      data.reset_labels()
   if 'self' in modes:
      results_dict['self'] = selftrain(data)
      data.reset_labels()

   data.reset_labels(False)
   return results_dict

def run_split(all_data, args):
   if not args.test_files: 
      views = all_data[0]
      logger.debug("Creating folds")
      ind = sample.SubsetIndicesCV(views[0], args.folds)
      for f in xrange(args.folds):
         validation = []
         training = []
         for view in views:
            validation.append(view.select_ref(ind, f))
            training.append(view.select_ref(ind, f, negate=True))
         
         logger.debug("Re-IDing data")
         for view in training:
            for i,row in enumerate(view):
               row[utils._ID] = i

         results = evaluate(training, validation, args)
         if f == 0:
            total_results = results
         else:
            for m in total_results:
               for v in total_results[m]:
                  for a in total_results[m][v]:
                     total_results[m][v][a]['Acc'] += results[m][v][a]['Acc']
                     total_results[m][v][a]['Recall'] += results[m][v][a]['Recall']
      for m in total_results:
         for v in total_results[m]:
            for a in total_results[m][v]:
               total_results[m][v][a]['Acc'] /= args.folds
               total_results[m][v][a]['Recall'] /= args.folds

   else:
      total_results = evaluate(*all_data, args=args)
   return total_results


def main():
   global v_start, v_end
   splits = [.00005, .0001, .0002, .0005, .001, .002, 
             .005, .01, .02, .05, .1, .2, .5]
   parser = ArgumentParser(description='Run Multiview Techniques', formatter_class=ArgumentDefaultsHelpFormatter)
   parser.add_argument('-s', '--split',  action='append', type=float, help='Set the percentage split of intially labeled data')
   parser.add_argument('-T', '--teacher_split', type=float, help='Set the percentage split of intially labeled data for the teacher')
   parser.add_argument('-b', '--bootstrap', action='store_true', help='Use the teacher to bootstrap labeled data')
   parser.add_argument('-N', default=1000, type=int, help='Set the size of N')
   parser.add_argument('-n', default=10, type=int, help='Set the size of n')
   parser.add_argument('-w', '--wsize', default=0, type=int, help='Set the window size')
   parser.add_argument('-S', '--save', help='Save the imputed/windowed data')
   parser.add_argument('-f', '--folds', default=10, type=int, help='Set the number of folds for cv')
   parser.add_argument('-v', '--verbose', default=0, action='count', help='Show verbose output')
   parser.add_argument('-l', '--log', action='store_true', help='Log to a file')
   parser.add_argument('-e', '--exclude-class', help='Remove instances of class EXCLUDE_CLASS')
   parser.add_argument('-i', '--impute', default='none', help='Set imputation technique', choices=['min', 'max', 'avg', 'none'])
   parser.add_argument('-a', '--algorithms', action='append', help='Sets the classifier choice', choices=['bayes', 'knn', 'svm', 'reg', 'tree'])
   parser.add_argument('-m', '--mode', action='append', help='Sets operating mode', choices=['baseline', 'maj', 'rand', 'cotrain', 'coem', 'manalign', 'teach', 'self'])
   parser.add_argument('-p', '--view-points', action='append', help='Sets the view start and end points. i.e. start,end')
   parser.add_argument('-c', '--class_var', help='Sets the class column')
   parser.add_argument('-t', '--test_files', help='Specify the test/validation files', nargs='*')
   parser.add_argument('data', help='Specify data files.', nargs='+')
   args = parser.parse_args()

   if args.verbose > 1:
      ch.setLevel(logging.DEBUG)
   elif args.verbose > 0:
      ch.setLevel(logging.INFO)
   if args.log:
      logger.addHandler(handler)

   logger.info("Starting co-training!")

   if not args.algorithms:
      args.algorithms = ['bayes', 'knn', 'tree', 'reg']
   if not args.mode:
      args.mode = ['baseline', 'maj', 'rand', 'cotrain', 'coem', 'manalign', 'teach']

   if args.view_points:
      try:
         vs = map(lambda x: int(x.split(',')[0]), args.view_points)
         ve = map(lambda x: int(x.split(',')[1]), args.view_points)
         v_start = vs
         v_end = ve
      except:
         logger.warn("Mal-formed view_points.  Reverting to defaults")


   all_data = utils.load_data(args.data, args.test_files)#, args.class_var)
   if not all_data[1]:
      all_data = (all_data[0],)

   all_data = utils.impute_data(all_data, args.impute)
   all_data = utils.window_data(all_data, args.wsize)

   if args.save:
      for i, data in enumerate(all_data):
         data.save('{0}({1})'.format(args.save, i))

   all_data = utils.set_class(all_data, args.class_var)

   if args.exclude_class:
      all_data = utils.filter_data(all_data, args.exclude_class) 

   logger.debug("IDing data")
   for i,row in enumerate(all_data[0]):
      row[utils._ID] = i

   all_data = utils.create_views(all_data, v_start, v_end)

   if args.split:
      splits = args.split
   total_results = {}
   for s in splits:
      logger.debug("Running Split {0}".format(s))
      args.split = s
      total_results[s] = run_split(all_data, args)


   print "Split Mode View Alg Acc Recall"
   for s in splits:   #Preserves the splits ordering
      for m in total_results[s]:
         for v in total_results[s][m]:
            for a in total_results[s][m][v]:
               acc = total_results[s][m][v][a]['Acc']
               rec = total_results[s][m][v][a]['Recall']
               print '{0} {1} {2} {3} {4} {5}'.format(s,m,v,a,acc,rec)

if __name__ == '__main__':
   import time
   sc = time.clock()
   sw = time.time()
   main()
   ec = time.clock()
   ew = time.time()
   logger.info('Computing Time: {0} s Wall Time: {1}'.format(ec-sc, ew-sw))
   import resource
   logger.info('Mem Used: {0} kb'.format(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss))

