#!/usr/bin/python
#!/usr/bin/env python
#*****************************************************************************#
#**
#**  Multiview Training and Testing Framework
#**
#**    Kyle D. Feuz, 2014
#**
#** Tools by the Center for Advanced Studies in Adaptive Systems at
#**  the School of Electrical Engineering and Computer Science at
#**  Washington State University
#** 
#** Copyright Washington State University, 2014
#** Copyright Kyle D. Feuz, 2014
#** 
#** Licenced under the Creative Commons (http://creativecommons.org):
#** Attribution-NonCommercial-ShareAlike 3.0 Unported
#** 
#** If this code is used for public research, any resulting publications need
#** to cite work done by Kyle D. Feuz at the Center for Advanced Study of 
#** Adaptive Systems (CASAS) at Washington State University.
#** 
#** Contact: Kyle D. Feuz (kyle.feuz@wsu.edu)
#** Contact: Diane J. Cook (cook@eecs.wsu.edu)
#*****************************************************************************#
import RandomLearner
import Orange
from Orange.data import Table, Domain, filter, Instance, Value
from Orange.data.imputation import ImputeTable
from Orange.feature import Descriptor, imputation, Type, Continuous
from Orange.evaluation import scoring
import numpy as np
import logging
try:
   from collections import Counter
except:
   from backport.Counter import Counter

logger = logging.getLogger(__name__)

_ID = Continuous("ID")
_new_id = Descriptor.new_meta_id()

def load_data(data_files, validation_files):#, class_var):
   """Load data from files"""

   data = []
   for fname in data_files:
      logger.info("Loading {0}".format(fname))
      data.append(Table(fname))

   for i in range(1, len(data)):
      data[0].extend(data[i])
   data = data[0]

   if not validation_files:
      validation_files = []

   validation = []
   for fname in validation_files:
      logger.info("Loading {0}".format(fname))
      validation.append(Table(fname))

   for i in range(1, len(validation)):
      validation[0].extend(validation[i])
   if not validation:
      validation.append(None)
   validation = validation[0]

   return data, validation
   #if validation:
   #   data = (data, validation)
   #else:
   #   data = (data,)

def set_class(data, class_var):

   if class_var:
      f,s = Descriptor.make(class_var, Type.Discrete)
      for d in data:
         d.pickClass(f)

   #Strip out other class labels and add ID
   domain = Domain(data[0].domain.features + [data[0].domain.class_var])
   domain.add_meta(_new_id, _ID)
   new_data = []
   for d in data:
      new_data.append(Table(domain, d))
   #if len(new_data) < 2:
   #   new_data.append(None)
   return new_data

def impute_data(data, mode):
   logger.info("Imputing data")
   if mode == "min":
      imputer = imputation.MinimalConstructor()
   elif mode == "max":
      imputer = imputation.MaximalConstructor()
   elif mode == "avg":
      imputer = imputation.AverageConstructor()
   else:
      imputer = None

   imputed = []
   if imputer != None:
      for d in data:
         imputed.append(ImputeTable(d, method=imputer))
   else:
      imputed = data
   return imputed

def window_data(data, wsize):
   windowed = []
   if wsize < 2:
      for d in data:
         windowed.append(Table(d))
      return windowed
   logger.info("Windowing data")
   for d in data:
      windowed.append(Table(d.domain))
      start = 0
      end = wsize
      while start < len(d):
         if end >= len(d):
            end = len(d)
         x,c,mc = d.get_items_ref(range(start,end)).to_numpy('a/cm')
         labels = []
         if mc == None:
            m = 0
         else:
            n,m = mc.shape
         labels.append(int(Counter(c).most_common(1)[0][0]))
         for i in xrange(m):
            labels.append(int(Counter(mc[:,i]).most_common(1)[0][0]))
         xm = np.mean(x, axis=0)
         x = Instance(d.domain, xm.tolist() + labels)
         #if int(x.get_classes()[0]) > 4:
         #   print labels[0]
         windowed[-1].append(x)
         start += wsize/2
         end += wsize/2
   return windowed

def filter_data(data, value='0', negate=True, ref=False):
   logger.info("Filtering data")
   f = filter.SameValue()
   f.value = Value(data[0].domain.class_var, value)
   f.position = len(data[0].domain.features)
   f.negate = negate
   filtered = []
   for d in data:
      if ref:
         filtered.append(d.filter_ref(f))
      else:
         filtered.append(d.filter(f))
   return filtered

def create_views(data, v_start, v_end):
   logger.info("Creating Views")
   view_domains = []
   for i in range(len(v_start)):
      logger.debug("View {0}: {1}-{2}".format(i, v_start[i], v_end[i]))
      view_domains.append(Domain(data[0].domain.features[v_start[i]:v_end[i]] + [data[0].domain.class_var]))
      view_domains[-1].add_meta(_new_id, _ID)

   views = []
   for d in data:
      views.append([])
      for vd in view_domains:
         views[-1].append(Table(vd, d))
   return views

def train(algs, train_views, view_number=None):
   classifiers = []
   for i,view in enumerate(train_views):
      if view_number != None:
         logger.info("Training View {0}".format(view_number))
      else:
         logger.info("Training View {0}".format(i))
      logger.debug("Number of Attributes: {0}".format(len(view.domain.features)))
      #for d in train_views[i]:
      #   if int(d.get_class()) > 4:
      #      print d

      #Append classifiers in order
      c = []
      for alg in algs:
         if alg == 'bayes':
            c.append(Orange.classification.bayes.NaiveLearner(view))
         if alg == 'knn':
            c.append(Orange.classification.knn.kNNLearner(view, k=7))
         if alg == 'svm':
            c.append(Orange.classification.svm.SVMLearnerEasy(view))
         if alg == 'reg':
            c.append(Orange.classification.logreg.LibLinearLogRegLearner(view))
         if alg == 'tree':
            c.append(Orange.classification.tree.TreeLearner(view, same_majority_pruning=1, m_pruning=2))
         if alg == 'maj':
            c.append(Orange.classification.majority.MajorityLearner(view))
         if alg == 'rand':
            c.append(RandomLearner.Learner(view))
      classifiers.append(c)
   return classifiers

def test(classifiers, test_views):
   #Evaluate classifiers   
   results = []
   for i in xrange(len(classifiers)):
      logger.info("Testing View {0}".format(i))
      logger.debug("Number of Attributes: {0}".format(len(test_views[i].domain.features)))
      results.append(Orange.evaluation.testing.test_on_data(classifiers[i],test_views[i]))

      #for d in test_views[i]:
      #   if int(d.get_class()) > 4:
      #      print d

   return results


def train_and_test(algs, train_views, test_views):
   classifiers = train(algs, train_views)
   results = test(classifiers, test_views)
   return results, classifiers

def get_acc(results):
   acc = []
   for result in results:
      acc.append(scoring.CA(result))
   return acc 

def get_recall(results):
   recall = []
   for result in results:
      recall.append(scoring.Avg_Recall(result))
   return recall

def show_results(results, names, cm=False):
   for i, result in enumerate(results):
      logger.info("Learner      CA      Avg_Recall")
      for j, name in enumerate(names):
         logger.info('{0}\t{1}\t{2}\n'.format(name, scoring.CA(result)[j], scoring.Avg_Recall(result)[j]))

      if cm:
         logger.info('Confusion Matrices')
         for j, name in enumerate(names):
            logger.info(name)
            cm = Orange.evaluation.scoring.confusion_matrices(result)[j]
            for row in cm:
               logger.info(row)

def train_test_report(algs, training, validation, 
                      show_cm=False, results_dict=None):
      
   classifiers = train(algs, training)
   results = test(classifiers, validation)
   acc = get_acc(results)
   rec = get_recall(results)

   if not results_dict:
      results_dict = {}
      for view in xrange(len(acc)):
         results_dict[view] = {}
         for i,alg in enumerate(algs):
            results_dict[view][alg] = {'Acc':0.0, 'Recall':0.0}

   for view in xrange(len(acc)):
      for i,alg in enumerate(algs):
         results_dict[view][alg]['Acc'] = acc[view][i]
         results_dict[view][alg]['Recall'] = rec[view][i]

   show_results(results, algs, show_cm)
   return results_dict
